# Copyright 2018 Descartes Labs.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
from collections import defaultdict, OrderedDict
import itertools
import json
import logging
import os
import sys
import time
from warnings import warn
from six.moves import zip_longest

import cloudpickle

from descarteslabs.client.auth import Auth
from descarteslabs.client.exceptions import NotFoundError
from descarteslabs.client.services.service import Service
from descarteslabs.client.services.storage import Storage
from descarteslabs.common.dotdict import DotDict, DotList


OFFSET_DEPRECATION_MESSAGE = (
    "Keyword arg `offset` has been deprecated and will be removed in "
    "future versions of the library. Use `continuation_token`."
)


class Tasks(Service):

    TASK_RESULT_BATCH_SIZE = 100
    RERUN_BATCH_SIZE = 200
    COMPLETION_POLL_INTERVAL_SECONDS = 5

    def __init__(self, url=None, auth=None):
        if auth is None:
            auth = Auth()

        if url is None:
            url = os.environ.get(
                "DESCARTESLABS_TASKS_URL",
                "https://platform.descarteslabs.com/tasks/v1"
            )

        super(Tasks, self).__init__(url, auth=auth)

    def create_namespace(self):
        """
        Creates a namespace for the user and sets up authentication within it
        from the current client id and secret. Must be called once per user
        before creating any tasks.

        :return: `True` if successful, `False` otherwise.
        """
        data = {
            'CLIENT_ID': self.auth.client_id,
            'CLIENT_SECRET': self.auth.client_secret
        }
        r = self.session.post('/namespaces/secrets/auth', json=data)
        return r.status_code == 201

    def new_group(
            self,
            function,
            container_image=None,
            name=None,
            cpus=1,
            memory='2Gi',
            maximum_concurrency=None,
            minimum_concurrency=None,
            minimum_seconds=None,
            task_timeout=1800,
    ):
        """
        Creates a new task group.

        :param function function: The function to be called in a task.
        :param str container_image: The location of a docker image to be used for
            the environment in which the function is executed.
        :param str name: An optional name used to later help identify the function.
        :param int cpus: The number of CPUs requested for a single task. A task
            might be throttled if it uses more CPU. Default: 1. Maximum: 32.
        :param str memory: The maximum memory requirement for a single task. If a
            task uses substantially more memory it will be killed. The value
            should be a string and can use postfixes such as Mi, Gi, MB, GB, etc
            (e.g. `"4Gi"`, `"500MB"`). If no unit is specified it is assumed to be
            in bytes. Default: 2Gi. Maximum: 64Gi.
        :param int maximum_concurrency: The maximum number of tasks to run in
            parallel. Default: 500. Maximum: 500. If you need higher concurrency
            contact your Descartes Labs customer success representative.
        :param int minimum_concurrency: The minimum number of tasks to run right
            away in parallel. Concurrency is usually scaled up slowly when
            submitting new tasks. Setting this can mean more immediate processing
            of this many newly submitted tasks. Note that setting this means the
            equivalent resources of this many permanently running tasks will be
            charged to your account while this group is active. Default: 0.
            Maximum: 4.
        :param int minimum_seconds: The number of seconds to wait for new tasks
            before scaling down concurrency, after a task is finished. Default: 0.
            Maximum: 600.
        :param int task_timeout: Maximum runtime for a single task in seconds. A
            task will be killed if it exceeds this limit. Default: 30 minutes.
            Minimum: 10 seconds. Maximum: 24 hours.

        :return: A dictionary representing the group created.
        """
        if container_image is None:
            container_image = "us.gcr.io/dl-ci-cd/images/tasks/public/alpha/py2/default:v2018.04.26"

        payload = {
            'image': container_image,
            'function': _serialize_function(function),
            'function_python_version': ".".join(str(part) for part in sys.version_info[:3]),
            'cpu': cpus,
            'mem': memory,
            'worker_timeout': task_timeout,
            'maximum_concurrency': maximum_concurrency,
            'minimum_concurrency': minimum_concurrency,
            'minimum_seconds': minimum_seconds,
        }

        if name is not None:
            payload['name'] = name

        r = self.session.post(
            "/groups",
            data=json.dumps(payload)
        )
        r.raise_for_status()
        return DotDict(r.json())

    def list_groups(
        self,
        status=None,
        created=None,
        updated=None,
        sort_field=None,
        sort_order="asc",
        limit=100,
        continuation_token=None,
    ):
        """
        Retrieves a limited list of task groups matching the given criteria.

        :param str status: Filter groups to this status.
            Allowed are ['running', 'terminated'].
        :param str created: Filter groups by creation date after this timestamp.
        :param str updated: Filter groups by updated date after this timestamp.
        :param str sort_field: The field to sort groups on. Allowed are
            ['created', 'updated'].
        :param str sort_order: Allowed are ['asc', 'desc']. Default: 'asc'.
        :param int limit: The number of results to get (max 1000 per page).
        :param str continuation_token: A string returned from a previous call to
            `list_groups()`, which you can use to get the next page of results.

        :return: A dictionary with two keys; `groups` containing the list of
            matching groups, `continuation_token` containting a string if there
            are further matching groups.
        """
        params = {'limit': limit}
        for field in ['status', 'created', 'updated', 'sort_field', 'sort_order', 'continuation_token']:
            if locals()[field] is not None:
                params[field] = locals()[field]
        r = self.session.get("/groups", params=params)
        r.raise_for_status()
        return DotDict(r.json())

    def iter_groups(
        self,
        status=None,
        created=None,
        updated=None,
        sort_field=None,
        sort_order="asc",
    ):
        """
        Iterates over all task groups matching the given criteria.

        :param str status: Filter groups to this status.
            Allowed are ['running', 'terminated'].
        :param str created: Filter groups by creation date after this timestamp.
        :param str updated: Filter groups by updated date after this timestamp.
        :param str sort_field: The field to sort groups on. Allowed are
            ['created', 'updated'].
        :param str sort_order: Allowed are ['asc', 'desc']. Default: 'asc'.
        :param int limit: The number of results to get (max 1000 per page).
        :param str continuation_token: A string returned from a previous call to
            `list_groups()`, which you can use to get the next page of results.

        :return: An iterator over matching task groups.
        """
        continuation_token = None
        while True:
            page = self.list_groups(status=status, created=created, updated=updated,
                                    sort_field=sort_field, sort_order=sort_order,
                                    continuation_token=continuation_token)
            for group in page.groups:
                yield group

            continuation_token = page.continuation_token
            if continuation_token is None:
                break

    def get_group(self, group_id):
        """
        Retrieves a single task group by id.

        :param str group_id: The group id.

        :return: A dictionary representing the task group.
        """
        r = self.session.get(
            "/groups/{}".format(group_id),
        )
        r.raise_for_status()
        return DotDict(r.json())

    get_group_by_id = get_group

    def get_group_by_name(self, name, status='running'):
        """
        Retrieves a single task group by name. Names are not unique; if there are
        multiple matches, returns the newest group.

        :param str group_id: The group name.
        :param str status: Only consider groups with this status.
            Allowed are ['running', 'terminated']. Default: 'running'.

        :return: A dictionary representing the task group, or `None` if no group
            with the given name exists.
        """
        groups = self.iter_groups(status=status, sort_field="created", sort_order="desc")
        for g in groups:
            if g.name == name:
                return g

    def terminate_group(self, group_id):
        """
        Terminates a task group by id. Once a group is terminated, no more tasks
        can be submitted to it and it stops using any resources. If the group
        with the given id is already terminated, nothing happens.

        :param str group_id: The group id.

        :return: A dictionary representing the terminated task group.
        """
        r = self.session.delete(
            "/groups/{uid}".format(uid=group_id)
        )
        r.raise_for_status()
        return DotDict(r.json())

    delete_group_by_id = terminate_group

    def wait_for_completion(self, group_id, show_progress=False):
        """
        Waits until all submitted tasks for a given group are completed.

        :param str group_id: The group id.
        :param bool show_progress: Whether to log progress information.
        """
        queue = self.get_group(group_id).queue
        while queue.pending > 0:
            completed = queue.failures + queue.successes
            if show_progress:
                logging.warning("Done with %i / %i tasks", completed, queue.pending + completed)
            time.sleep(self.COMPLETION_POLL_INTERVAL_SECONDS)
            queue = self.get_group(group_id).queue

    def new_task(self, group_id, arguments=None, parameters=None,
                 labels=None, retry_count=0):
        """
        Submits a new task to a group. All positional and keyword arguments
        to the group's function must be JSON-serializable (i.e., booleans,
        numbers, strings, lists, dictionaries).

        :param str group_id: The group id to submit to.
        :param list arguments: The positional arguments to call the group's
            function with.
        :param dict parameters: The keyword arguments to call the group's
            function with.
        :param list labels: An optional list of labels to attach to the
            task. Task results can later be filtered by these labels.
        :param int retry_count: Number of times to retry the task if it
            fails (maximum 5).

        :return: A dictionary with one key `tasks` containing a list with
            one element representing the submitted task.
        """
        return self.new_tasks(
            group_id,
            list_of_arguments=[arguments or []],
            list_of_parameters=[parameters or {}],
            list_of_labels=[labels],
            retry_count=retry_count
        )

    def new_tasks(self, group_id, list_of_arguments=None,
                  list_of_parameters=None, list_of_labels=None,
                  retry_count=0):
        """
        Submits multiple tasks to a group. All positional and keyword arguments
        to the group's function must be JSON-serializable (i.e., booleans,
        numbers, strings, lists, dictionaries).

        :param str group_id: The group id to submit to.
        :param list(list) arguments: The positional arguments to call the
            group's function with, for each task.
        :param list(dict) parameters: The keyword arguments to call the group's
            function with, for each task.
        :param list(list) labels: An optional list of labels to attach, for each
            task. Task results can later be filtered by these labels.
        :param int retry_count: Number of times to retry the tasks if they
            fails (maximum 5).

        :return: A dictionary with one key `tasks` containing a list of
            dictionaries representing the submitted tasks.
        """
        list_of_arguments = list_of_arguments if \
            list_of_arguments is not None else [[]]
        list_of_parameters = list_of_parameters if \
            list_of_parameters is not None else [{}]
        list_of_labels = list_of_labels if \
            list_of_labels is not None else [None]
        msgs = []
        for args, kwargs, labels in zip_longest(list_of_arguments, list_of_parameters, list_of_labels, fillvalue=None):
            args = args or []
            params = kwargs or {}
            attributes = {'labels': labels} if labels else {}
            msg = {
                'arguments': args,
                'parameters': params,
                'attributes': attributes,
                'retry_count': retry_count,
            }
            msgs.append(msg)

        r = self.session.post(
            "/groups/{group_id}/tasks".format(group_id=group_id),
            json={
                'tasks': msgs
            }
        )
        r.raise_for_status()
        return DotDict(r.json())

    def get_task_result(self, group_id, task_id, include=None):
        """
        Retrieves a single task result.

        :param str group_id: The group to get task results from.
        :param str task_id: Specific ID of task to retrieve.
        :param list(str) include: Extra fields to include in the task results.
            Allowed values are ['arguments', 'stacktrace', 'result', 'logs',
            'result_url', 'logs_url'].

        :return: A dictionary representing the task result.
        """
        params = {'include': include} if include is not None else {}
        r = self.session.get(
            '/groups/{uid}/tasks/{task_uid}/results'.format(
                uid=group_id, task_uid=task_id
            ),
            params=params,
        )
        r.raise_for_status()
        result = r.json()
        if 'result' in result:
            result['result'] = base64.b64decode(result['result'])
        return DotDict(result)

    def get_task_result_batch(self, group_id, task_ids, include=None):
        """
        Retrieves a multiple task results by id.

        :param str group_id: The group to get task results from.
        :param list(str) task_ids: A list of task ids to retrieve, maximum 500.
        :param list(str) include: Extra fields to include in the task results.
            Allowed values are ['arguments', 'stacktrace', 'result_url', 'logs_url'].

        :return: A dictionary with a key `results` containing the list of
            matching results. Results are in the order of the ids provided.
            Unknown ids are ignored.
        """
        data = {'ids': task_ids}
        if include is not None:
            data['include'] = include
        r = self.session.post(
            '/groups/{uid}/results/batch'.format(uid=group_id),
            json=data,
        )
        r.raise_for_status()
        results = r.json()
        return DotDict(results)

    def list_task_results(
            self,
            group_id,
            limit=TASK_RESULT_BATCH_SIZE,
            offset=None,
            status=None,
            failure_type=None,
            updated=None,
            created=None,
            webhook=None,
            labels=None,
            include=None,
            sort_field='created',
            sort_order='asc',
            continuation_token=None,
    ):
        """
        Retrieves a limited list of task results matching the given criteria.

        :param str group_id: The group to get task results from.
        :param int limit: The number of results to get (max 1000 per page).
        :param int offset: Where to start when getting task results
            (deprecated; use continuation_token).
        :param str status: Filter tasks to this status.
            Allowed are ['FAILURE', 'SUCCESS'].
        :param str failure_type: Filter tasks to this type of failure.
            Allowed are ['exception', 'oom', 'timeout', 'internal', 'unknown'].
        :param str updated: Filter tasks by updated date after this timestamp.
        :param str created: Filter tasks by creation date after this timestamp.
        :param str webhook: Filter by the webhook uid which spawned the task.
        :param list(str) labels: Labels that must be present in tasks labels list.
        :param list(str) include: Extra fields to include in the task results.
            Allowed values are ['arguments', 'stacktrace', 'result_url', 'logs_url'].
        :param str sort_field: The field to sort results on. Allowed are
            ['created', 'runtime', 'peak_memory_usage']. Default: 'created'.
        :param str sort_order: Allowed are ['asc', 'desc']. Default: 'asc'.
        :param str continuation_token: A string returned from a previous call to
            `list_task_results()`, which you can use to get the next page of results.

        :return: A dictionary with two keys; `results` containing the list of
            matching results, `continuation_token` containting a string if there
            are further matching results.
        """
        if offset is not None:
            warn(OFFSET_DEPRECATION_MESSAGE, DeprecationWarning)

        params = {'limit': limit}
        for field in ['offset', 'status', 'failure_type', 'updated', 'created', 'webhook',
                      'labels', 'include', 'sort_field', 'sort_order', 'continuation_token']:
            if locals()[field] is not None:
                params[field] = locals()[field]
        r = self.session.get(
            '/groups/{uid}/results'.format(uid=group_id),
            params=params
        )
        r.raise_for_status()
        return DotDict(r.json())

    get_task_results = list_task_results

    def iter_task_results(
            self,
            group_id,
            status=None,
            failure_type=None,
            updated=None,
            created=None,
            webhook=None,
            labels=None,
            include=None,
            sort_field='created',
            sort_order='asc',
    ):
        """
        Iterates over all task results matching the given criteria.

        :param str group_id: The group to get task results from.
        :param str status: Filter tasks to this status.
            Allowed are ['FAILURE', 'SUCCESS'].
        :param str failure_type: Filter tasks to this type of failure.
            Allowed are ['exception', 'oom', 'timeout', 'internal', 'unknown'].
        :param str updated: Filter tasks by updated date after this timestamp.
        :param str created: Filter tasks by creation date after this timestamp.
        :param str webhook: Filter by the webhook uid which spawned the task.
        :param list(str) include: Extra fields to include in the task results.
            Allowed values are ['arguments', 'stacktrace', 'result_url', 'logs_url'].
        :param list(str) labels: Labels that must be present in tasks labels list.
        :param str sort_field: The field to sort results on. Allowed are
            ['created', 'runtime', 'peak_memory_usage']. Default: 'created'.
        :param str sort_order: Allowed are ['asc', 'desc']. Default: 'asc'.

        :return: An iterator over matching task results.
        """
        params = {}
        for field in ['status', 'failure_type', 'updated', 'created', 'webhook', 'labels', 'include']:
            if locals()[field] is not None:
                params[field] = locals()[field]

        continuation_token = None
        while True:
            page = self.get_task_results(group_id, continuation_token=continuation_token, **params)
            for result in page.results:
                yield result

            continuation_token = page.continuation_token
            if continuation_token is None:
                break

    def rerun_failed_tasks(self, group_id, retry_count=0):
        """
        Submits all failed tasks for a rerun, except for tasks that had an
        out-of-memory failure. These tasks will be run again with the same
        arguments as before.

        Tasks that are currently already being rerun will be ignored.

        :param str group_id: The group in which to rerun tasks.
        :param int retry_count: Number of times to retry a task if it fails
                                (maximum 5)

        :return: A list of dictionaries representing the tasks that have been submitted.
        """
        rerun_tasks = []
        for failure_type in ['exception', 'timeout', 'internal', 'unknown']:
            rerun_tasks += self.rerun_matching_tasks(group_id, failure_type=failure_type, retry_count=retry_count)
        return rerun_tasks

    def rerun_matching_tasks(
            self,
            group_id,
            status=None,
            failure_type=None,
            updated=None,
            created=None,
            webhook=None,
            labels=None,
            retry_count=0
    ):
        """
        Submits all completed tasks matching the given search arguments for a rerun.
        These tasks will be run again with the same arguments as before.

        Tasks that are currently already being rerun will be ignored.

        :param str group_id: The group in which to rerun tasks.
        :param str status: Filter tasks to this status.
            Allowed are ['FAILURE', 'SUCCESS'].
        :param str failure_type: Filter tasks to this type of failure.
            Allowed are ['exception', 'oom', 'timeout', 'internal', 'unknown'].
        :param str updated: Filter tasks by updated date after this timestamp.
        :param str created: Filter tasks by creation date after this timestamp.
        :param str webhook: Filter by the webhook uid which spawned the task.
        :param list(str) labels: Labels that must be present in tasks labels list.
        :param int retry_count: Number of times to retry a task if it fails
                                (maximum 5)

        :return: A list of dictionaries representing the tasks that have been submitted.
        """
        results = self.iter_task_results(group_id, status=status, failure_type=failure_type, updated=updated,
                                         created=created, webhook=webhook, labels=labels)
        return self.rerun_tasks(group_id, (t.id for t in results), retry_count=retry_count)

    def rerun_tasks(self, group_id, task_id_iterable, retry_count=0):
        """
        Submits a list of completed tasks specified by ids for a rerun. The completed tasks
        with the given ids will be run again with the same arguments as before.

        Tasks that are currently already being rerun will be ignored. Unknown or invalid
        task ids will be ignored.

        :param str group_id: The group in which to rerun tasks.
        :param iterable(str) task_id_iterable: An iterable of the task ids to be rerun.
        :param int retry_count: Number of times to retry a task if it fails
                                (maximum 5)

        :return: A list of dictionaries representing the tasks that have been submitted.
        """
        rerun = []
        task_ids = list(itertools.islice(task_id_iterable, self.RERUN_BATCH_SIZE))
        while task_ids:
            r = self.session.post(
                "/groups/{group_id}/tasks/rerun".format(group_id=group_id),
                json={
                    'task_ids': task_ids,
                    'retry_count': retry_count,
                }
            )
            r.raise_for_status()
            rerun += r.json()['tasks']
            task_ids = list(itertools.islice(task_id_iterable, self.RERUN_BATCH_SIZE))
        return DotList(rerun)

    def create_function(self, f, image=None, name=None, cpus=1,
                        memory='2Gi',
                        maximum_concurrency=None,
                        minimum_concurrency=None,
                        minimum_seconds=None,
                        task_timeout=1800,
                        retry_count=0,
                        ):
        """
        Creates a new task group from a function and returns an asynchronous
        function that can be called to submit tasks to the group.

        :param function f: The function to be called in a task.
        :param str image: The location of a docker image to be used for the
            environment in which the function is executed.
        :param str name: An optional name used to later help identify the function.
        :param int cpus: The number of CPUs requested for a single task. A task
            might be throttled if it uses more CPU. Default: 1. Maximum: 32.
        :param str memory: The maximum memory requirement for a single task. If a
            task uses substantially more memory it will be killed. The value
            should be a string and can use postfixes such as Mi, Gi, MB, GB, etc
            (e.g. `"4Gi"`, `"500MB"`). If no unit is specified it is assumed to be
            in bytes. Default: 2Gi. Maximum: 64Gi.
        :param int maximum_concurrency: The maximum number of tasks to run in
            parallel. Default: 500. Maximum: 500. If you need higher concurrency
            contact your Descartes Labs customer success representative.
        :param int minimum_concurrency: The minimum number of tasks to run right
            away in parallel. Concurrency is usually scaled up slowly when
            submitting new tasks. Setting this can mean more immediate processing
            of this many newly submitted tasks. Note that setting this means the
            equivalent resources of this many permanently running tasks will be
            charged to your account while this group is active. Default: 0.
            Maximum: 4.
        :param int minimum_seconds: The number of seconds to wait for new tasks
            before scaling down concurrency, after a task is finished. Default: 0.
            Maximum: 600.
        :param int task_timeout: Maximum runtime for a single task in seconds. A
            task will be killed if it exceeds this limit. Default: 30 minutes.
            Minimum: 10 seconds. Maximum: 24 hours.
        :param int retry_count: Number of times to retry a task if it fails
            Default: 0. Maximum: 5.

        :return: A :class:`CloudFunction`.
        """
        group_info = self.new_group(
            f, container_image=image, name=name,
            cpus=cpus, memory=memory,
            maximum_concurrency=maximum_concurrency,
            minimum_concurrency=minimum_concurrency,
            minimum_seconds=minimum_seconds,
            task_timeout=task_timeout,
        )

        return CloudFunction(group_info.id, name=name, client=self, retry_count=retry_count)

    def get_function(self, name):
        """
        Gets an asynchronous function by name (the last function created with that
        name).

        :param str name: The name of the function to lookup.

        :return: A :class:`CloudFunction`, or `None` if no function with the
            given name exists.
        """
        group_info = self.get_group_by_name(name)
        if group_info is None:
            return None

        return CloudFunction(group_info.id, name=name, client=self)

    def create_or_get_function(self, f, image=None, name=None, cpus=1,
                               memory='2Gi',
                               maximum_concurrency=None,
                               minimum_concurrency=None,
                               minimum_seconds=None,
                               task_timeout=1800,
                               retry_count=0,
                               ):
        """
        Creates or gets an asynchronous function. If a task group with the given
        name exists, returns an asynchronous function for the newest existing
        group with that. Otherwise creates a new task group.

        :param function f: The function to be called in a task.
        :param str image: The location of a docker image to be used for the
            environment in which the function is executed.
        :param str name: An optional name used to later help identify the function.
        :param int cpus: The number of CPUs requested for a single task. A task
            might be throttled if it uses more CPU. Default: 1. Maximum: 32.
        :param str memory: The maximum memory requirement for a single task. If a
            task uses substantially more memory it will be killed. The value
            should be a string and can use postfixes such as Mi, Gi, MB, GB, etc
            (e.g. `"4Gi"`, `"500MB"`). If no unit is specified it is assumed to be
            in bytes. Default: 2Gi. Maximum: 64Gi.
        :param int maximum_concurrency: The maximum number of tasks to run in
            parallel. Default: 500. Maximum: 500. If you need higher concurrency
            contact your Descartes Labs customer success representative.
        :param int minimum_concurrency: The minimum number of tasks to run right
            away in parallel. Concurrency is usually scaled up slowly when
            submitting new tasks. Setting this can mean more immediate processing
            of this many newly submitted tasks. Note that setting this means the
            equivalent resources of this many permanently running tasks will be
            charged to your account while this group is active. Default: 0.
            Maximum: 4.
        :param int minimum_seconds: The number of seconds to wait for new tasks
            before scaling down concurrency, after a task is finished. Default: 0.
            Maximum: 600.
        :param int task_timeout: Maximum runtime for a single task in seconds. A
            task will be killed if it exceeds this limit. Default: 30 minutes.
            Minimum: 10 seconds. Maximum: 24 hours.
        :param int retry_count: Number of times to retry a task if it fails
            Default: 0. Maximum: 5.

        :return: A :class:`CloudFunction`.
        """
        if name:
            cached = self.get_function(name)
            if cached is not None:
                return cached
        return self.create_function(
            f, image=image, name=name, cpus=cpus,
            memory=memory,
            maximum_concurrency=maximum_concurrency,
            minimum_concurrency=minimum_concurrency,
            minimum_seconds=minimum_seconds,
            task_timeout=task_timeout,
            retry_count=retry_count,
        )

    def create_webhook(self, group_id, name=None, label_path=None, label_separator=None):
        data = {}
        if name is not None:
            data['name'] = name
        if label_path is not None:
            data['label_path'] = label_path
        if label_separator is not None:
            data['label_separator'] = label_separator
        r = self.session.post(
            '/groups/{group_id}/webhooks'.format(group_id=group_id),
            json=data
        )
        r.raise_for_status()
        return DotDict(r.json())

    def list_webhooks(self, group_id):
        r = self.session.get(
            '/groups/{group_id}/webhooks'.format(group_id=group_id),
        )
        r.raise_for_status()
        return DotDict(r.json())

    get_webhooks = list_webhooks

    def get_webhook(self, group_id, webhook_id):
        r = self.session.get(
            '/groups/{group_id}/webhooks/{webhook_id}'.format(
                group_id=group_id,
                webhook_id=webhook_id,
            ),
        )
        r.raise_for_status()
        return DotDict(r.json())

    def delete_webhook(self, group_id, webhook_id):
        r = self.session.delete(
            '/groups/{group_id}/webhooks/{webhook_id}'.format(
                group_id=group_id,
                webhook_id=webhook_id,
            ),
        )
        r.raise_for_status()
        return True


AsyncTasks = Tasks


class TransientResultException(Exception):
    """
    Raised when attempting to access results for a task that hasn't
    completed.
    """

    def __init__(self, message):
        super(Exception, self).__init__(message)


class ResultType(object):
    """
    Possible types of return values for a function.
    """
    JSON = "json"
    LEGACY_PICKLE = "pickle"


class FutureTask(dict):
    """
    A submitted task which may or may not have completed yet. Accessing any
    attributes only available on a completed task (for example `result`)
    blocks until the task completes.
    """

    COMPLETION_POLL_INTERVAL_SECONDS = 3

    def __init__(self, guid, tuid, client=None, args=None, kwargs=None):
        self.guid = guid
        self.tuid = tuid
        self.client = client
        self.args = args
        self.kwargs = kwargs
        self._is_return_value_loaded = False
        self._return_value = None
        self._task_result = None
        self._is_log_loaded = False
        self._log = None
        self._json_arguments = None

    def get_result(self, wait=False, timeout=None):
        """
        Attempt to load the result for this task. After returning from this
        method without an exception raised, the return value for the task is
        available through the `result` property.

        :param bool wait: Whether to wait for the task to complete or raise
            a :class:`TransientResultException` if the task hasnt completed
            yet.
        :param int timeout: How long to wait for the task to complete, or
            `None` to wait indefinitely.
        """
        start = time.time()
        if self._task_result is None:
            while timeout is None or \
                    (time.time() - start) < timeout:
                try:
                    self._task_result = self.client.get_task_result(
                        self.guid, self.tuid, include=['stacktrace'])
                except NotFoundError:
                    if not wait:
                        raise TransientResultException("Result not yet ready")
                else:
                    break
                time.sleep(self.COMPLETION_POLL_INTERVAL_SECONDS)
            else:
                raise RuntimeError("Timeout exceeded")
        return

    def result_attribute(self, attribute_name, default=None):
        if self._task_result is None:
            self.get_result(wait=True)
        return self._task_result.get(attribute_name, default)

    @property
    def result(self):
        """
        :return: The return value of the function for this completed task.
        """
        if self._task_result is None:
            self.get_result(wait=True)

        if not self.is_success:
            return None

        if not self._is_return_value_loaded:
            return_value = Storage().get(self._task_result.result_key, storage_type='result')
            result_type = self._task_result.get('result_type', ResultType.LEGACY_PICKLE)
            if result_type == ResultType.LEGACY_PICKLE:
                return_value = cloudpickle.loads(return_value)
                if isinstance(return_value, dict):
                    # For backwards-compatibility reasons, for legacy pickles the result is
                    # wrapped in a dictionary.
                    self._return_value = return_value['result']
                else:
                    self._return_value = return_value
            elif result_type == ResultType.JSON:
                self._return_value = json.loads(return_value.decode('utf-8'))
            else:
                raise RuntimeError("Unknown result type: %s - update your tasks client")
            self._is_return_value_loaded = True

        return self._return_value

    @property
    def log(self):
        """
        :return: The log output for this completed task.
        """
        if self._task_result is None:
            self.get_result(wait=True)
        if not self._is_log_loaded and self._task_result.get('log_size_bytes', 1) > 0:
            try:
                self._log = Storage().get(self._task_result.result_key, storage_type='logs')
            except NotFoundError:
                self._log = None
            self._is_log_loaded = True
        return self._log

    @property
    def peak_memory_usage(self):
        """
        :return: The peak memory usage for this completed task, in bytes.
        """
        return self.result_attribute('peak_memory_usage')

    @property
    def runtime(self):
        """
        :return: The time spent executing the function for this task, in seconds.
        """
        return self.result_attribute('runtime')

    @property
    def status(self):
        """
        :return: The status (`SUCCESS` or `FAILURE`) for this completed task.
        """
        return self.result_attribute('status')

    @property
    def is_success(self):
        """
        :return: Whether this task succeeded.
        """
        return self.status == "SUCCESS"

    @property
    def exception_name(self):
        """
        :return: The name of the exception raised during the function execution,
            if any.
        """
        return self.result_attribute('exception_name')

    exception = exception_name

    @property
    def stacktrace(self):
        """
        :return: The stacktrace of the exception raised during the function
            execution, if any.
        """
        return self.result_attribute('stacktrace')

    traceback = stacktrace

    @property
    def failure_type(self):
        """
        :return: The type of failure if this task did not succeed.
        """
        return self.result_attribute('failure_type')

    def __eq__(self, other):
        return self.guid == other.guid and self.tuid == other.tuid

    def __repr__(self):
        s = "Task\n"
        if self._task_result is None:
            s += "\tStatus: Pending\n"
        else:
            s += "\tStatus: {}\n".format(self.status)
            s += "\tMemory usage (MiB): {:.2f}\n".format(
                self.peak_memory_usage / (1024 * 1024.)
            )
            s += "\tRuntime (s): {}\n".format(self.runtime)
        return s


class CloudFunction(object):
    """
    Represents the asynchronous function of a task group. When called, new
    tasks are submitted to the group with the positional and keyword arguments
    given. A `map()` method allows submitting multiple tasks more efficiently
    than making individual function calls.
    """
    TASK_SUBMIT_SIZE = 100

    def __init__(self, group_id, name=None, client=None, retry_count=0):
        self.group_id = group_id
        self.name = name
        self.client = client
        self.retry_count = retry_count

    def __call__(self, *args, **kwargs):
        """
        Submits a task calling the function with the given positional and
        keyword arguments.

        All positional and keyword arguments must be JSON-serializable (i.e.,
        booleans, numbers, strings, lists, dictionaries).

        :return: A :class:`FutureTask` for the submitted task.
        """
        tasks = self.client.new_task(
            self.group_id,
            arguments=args,
            parameters=kwargs,
            retry_count=self.retry_count,
        )
        task_info = tasks.tasks[0]
        return FutureTask(self.group_id, task_info.id, client=self.client, args=args, kwargs=kwargs)

    def map(self, args, *iterargs):
        """
        Submits multiple tasks efficiently with positional argument to each function
        call, mimicking the behaviour of the builtin `map()` function. When
        submitting multiple tasks this is preferred over calling the function
        repeatedly.

        All positional arguments must be JSON-serializable (i.e., booleans, numbers,
        strings, lists, dictionaries).

        :param iterable args: An iterable of arguments. A task will be submitted
            with each element of the iterable as the first positional argument
            to the function.
        :param list(iterable) iterargs: If additional iterable arguments are
            passed, the function must take that many arguments and is applied
            to the items from all iterables in parallel (mimicking builtin
            `map()` behaviour).

        :return: A list of :class:`FutureTask` for all submitted tasks.
        """
        arguments = zip_longest(args, *iterargs)

        futures = []
        batch = list(itertools.islice(arguments, self.TASK_SUBMIT_SIZE))
        while batch:
            tasks_info = self.client.new_tasks(
                self.group_id,
                list_of_arguments=batch,
                retry_count=self.retry_count,
            )
            futures += [
                FutureTask(self.group_id, task_info.id, client=self.client, args=task_args)
                for task_info, task_args in zip(tasks_info.tasks, batch)
            ]
            batch = list(itertools.islice(arguments, self.TASK_SUBMIT_SIZE))

        return futures

    def wait_for_completion(self, show_progress=False):
        """
        Waits until all tasks submitted through this function are completed.

        :param bool show_progress: Whether to log progress information.
        """
        self.client.wait_for_completion(self.group_id, show_progress=show_progress)


def as_completed(tasks, show_progress=True):
    """
    Yields completed tasks from the list of given tasks as they become
    available, finishing when all given tasks have been completed.

    If you don't care about the particular results of the tasks and only
    want to wait for all tasks to complete, use
    :meth:`wait_for_completion <CloudFunction>`.

    :param list tasks: List of :class:`FutureTask` objects.
    :param bool show_progress: Whether to log progress information.
    """
    total_tasks = len(tasks)
    remaining = OrderedDict(((t.guid, t.tuid), t) for t in tasks)

    while len(remaining) > 0:
        batch = itertools.islice(remaining.values(), Tasks.TASK_RESULT_BATCH_SIZE)
        by_group = defaultdict(list)
        for task in batch:
            by_group[task.guid].append(task)

        for group_id, group_tasks in by_group.items():
            client = group_tasks[0].client
            task_ids = [task.tuid for task in group_tasks]
            try:
                results = client.get_task_result_batch(group_id, task_ids, include=['stacktrace'])
            except BaseException:
                logging.warning("Task retrieval for group %s failed with fatal error", group_id, exc_info=True)
            else:
                for result in results['results']:
                    task = remaining.pop((group_id, result.id))
                    task._task_result = result
                    yield task

        if show_progress:
            logging.warning("Done with %i / %i tasks", total_tasks - len(remaining), total_tasks)
        time.sleep(Tasks.COMPLETION_POLL_INTERVAL_SECONDS)


def _serialize_function(function):
    # Note; In Py3 cloudpickle and base64 handle bytes objects only, so we need to
    # decode it into a string to be able to json dump it again later.
    encoded_bytes = base64.b64encode(cloudpickle.dumps(function))
    return encoded_bytes.decode('ascii')



# Copyright 2018 Descartes Labs.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class Expression(object):
    def __and__(self, other):
        return AndExpression([self]) & other

    def __or__(self, other):
        return OrExpression([self]) | other

    def __rand__(self, other):
        return AndExpression([other]) & self

    def __ror__(self, other):
        return OrExpression([other]) | self


class EqExpression(Expression):
    def __init__(self, name, value):
        self.name = name
        self.value = value

    def serialize(self):
        return {'eq': {self.name: self.value}}


class NeExpression(Expression):
    def __init__(self, name, value):
        self.name = name
        self.value = value

    def serialize(self):
        return {'ne': {self.name: self.value}}


class RangeExpression(Expression):
    def __init__(self, name, parts):
        self.name = name
        self.parts = parts

    def serialize(self):
        return {'range': {self.name: self.parts}}


class AndExpression(object):
    def __init__(self, parts):
        self.parts = parts

    def __and__(self, other):
        if isinstance(other, AndExpression):
            self.parts.extend(other.parts)
            return self
        if isinstance(other, (OrExpression, Expression)):
            self.parts.append(other)
            return self
        else:
            raise Exception("Invalid sub-expression")

    __rand__ = __and__

    def __repr__(self):
        return '<AndExpression {}>'.format(self.parts)

    def serialize(self):
        return {'and': [x.serialize() for x in self.parts]}


class OrExpression(object):
    def __init__(self, parts):
        self.parts = parts

    def __or__(self, other):
        if isinstance(other, OrExpression):
            self.parts.extend(other.parts)
            return self
        if isinstance(other, (AndExpression, Expression)):
            self.parts.append(other)
            return self
        else:
            raise Exception("Invalid sub-expression")

    __ror__ = __or__

    def __repr__(self):
        return '<OrExpression {}>'.format(self.parts)

    def serialize(self):
        return {'or': [x.serialize() for x in self.parts]}


def range_expr(op):
    def f(self, other):
        # This is a hack to support compound comparisons
        # such as 10 < a < 20
        self.parts[op] = other
        return RangeExpression(self.name, self.parts.copy())
    return f


class Property(object):
    def __init__(self, name, parts=None):
        self.name = name
        self.parts = parts or {}

    __ge__ = range_expr('gte')
    __gt__ = range_expr('gt')
    __le__ = range_expr('lte')
    __lt__ = range_expr('lt')

    def __eq__(self, other):
        return EqExpression(self.name, other)

    def __ne__(self, other):
        return NeExpression(self.name, other)

    def __repr__(self):
        return '<Property {}>'.format(self.name)


class Properties(object):
    def __init__(self, *args):
        self.props = args

    def __getattr__(self, attr):
        if attr in self.props:
            return Property(attr)

        raise AttributeError(
            "'Properties' object has no attribute '{}'".format(attr))


GLOBAL_PROPERTIES = Properties(
    "absolute_orbit", "acquired", "archived", "area", "azimuth_angle",
    "bright_fraction", "cirrus_fraction", "cloud_fraction", "cloud_fraction_0",
    "cloud_fraction_1", "descartes_version", "earth_sun_distance",
    "fill_fraction", "geolocation_accuracy", "opaque_fraction", "product", "processed",
    "published", "relative_orbit", "roll_angle", "sat_id", "solar_azimuth_angle",
    "solar_azimuth_angle_0", "solar_elevation_angle", "solar_elevation_angle_0",
    "sw_version", "terrain_correction", "tile_id", "view_angle", "view_angle_0",
    "view_angle_1"
)



# Copyright 2018 Descartes Labs.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import struct
from io import BytesIO
import logging

import six

from descarteslabs.client.addons import ThirdParty, blosc, concurrent, numpy as np
from descarteslabs.client.auth import Auth
from descarteslabs.client.services.places import Places
from descarteslabs.client.services.service.service import Service
from descarteslabs.client.exceptions import ServerError
from descarteslabs.common.dotdict import DotDict


DEFAULT_MAX_WORKERS = 8


def as_json_string(str_or_dict):
    if not str_or_dict:
        return str_or_dict
    elif isinstance(str_or_dict, dict):
        return json.dumps(str_or_dict)
    else:
        return str_or_dict


def read_blosc_buffer(data):
    header = data.read(16)

    _, size, _, compressed_size = struct.unpack('<IIII', header)
    body = data.read(compressed_size - 16)

    return size, header + body


def read_blosc_array(metadata, data):
    output = np.empty(metadata['shape'], dtype=np.dtype(metadata['dtype']))
    ptr = output.__array_interface__['data'][0]

    for _ in metadata['chunks']:
        raw_size, buffer = read_blosc_buffer(data)
        blosc.decompress_ptr(buffer, ptr)
        ptr += raw_size

    bytes_received = ptr - output.__array_interface__['data'][0]

    if bytes_received != output.nbytes:
        raise ServerError("Did not receive complete array (got {}, expected {})".format(
            bytes_received, output.nbytes))

    return output


def read_blosc_string(metadata, data):
    output = b''

    for _ in metadata['chunks']:
        _, buffer = read_blosc_buffer(data)
        output += blosc.decompress(buffer)

    return output


class Raster(Service):
    """Raster"""
    TIMEOUT = (9.5, 300)

    def __init__(self, url=None, auth=None):
        """The parent Service class implements authentication and exponential
        backoff/retry. Override the url parameter to use a different instance
        of the backing service.
        """
        if auth is None:
            auth = Auth()

        if url is None:
            url = os.environ.get("DESCARTESLABS_RASTER_URL", "https://platform.descarteslabs.com/raster/v1")

        super(Raster, self).__init__(url, auth=auth)

    def dltiles_from_shape(self, resolution, tilesize, pad, shape):
        """
        Return a feature collection of DLTile GeoJSONs that intersect
        a GeoJSON Geometry `shape`.

        :param float resolution: Resolution of DLTile
        :param int tilesize: Number of valid pixels per DLTile
        :param int pad: Number of ghost pixels per DLTile (overlap among tiles)
        :param str shape: A GeoJSON geometry specifying a shape over
            which to intersect DLTiles.

        :return: GeoJSON FeatureCollection of intersecting DLTile geometries.

        Example::

            >>> from descarteslabs.client.services import Raster, Places
            >>> iowa = Places().shape("north-america_united-states_iowa")
            >>> tiles = Raster().dltiles_from_shape(30.0, 2048, 16, iowa)
            >>> tiles['features'][0]
            {
              'geometry': {
                'coordinates': [
                  [
                    [-96.81264975325402, 41.045203319986356],
                    [-96.07101667769108, 41.02873098016475],
                    [-96.04576296033223, 41.59007261142797],
                    [-96.79377566762066, 41.60687154946031],
                    ...
                  ]
                ],
                'type': 'Polygon'
              },
              'properties': {
                'cs_code': 'EPSG:32614',
                'key': '2048:16:30.0:14:3:74',
                'outputBounds': [683840.0, 4546080.0, 746240.0, 4608480.0],
                'pad': 16,
                'resolution': 30.0,
                'ti': 3,
                'tilesize': 2048,
                'tj': 74,
                'zone': 14
              },
              'type': 'Feature'
            }
        """

        shape = as_json_string(shape)
        params = {
            'resolution': resolution,
            'tilesize': tilesize,
            'pad': pad,
            'shape': shape,
        }

        r = self.session.post('/dlkeys/from_shape', json=params)
        return DotDict(r.json())

    def dltile_from_latlon(self, lat, lon, resolution, tilesize, pad):
        """
        Return a DLTile GeoJSON Feature that covers a latitude/longitude

        :param float lat: Requested latitude
        :param float lon: Requested longitude
        :param float resolution: Resolution of DLTile
        :param int tilesize: Number of valid pixels per DLTile
        :param int pad: Number of ghost pixels per DLTile (overlap among tiles)

        :return: A DLTile GeoJSON Feature

        Example::

            >>> from descarteslabs.client.services import Raster
            >>> Raster().dltile_from_latlon(45, 60, 15.0, 1024, 16)
            {
              'geometry': {
                'coordinates': [
                  [
                    [59.88428127486419, 44.89851158847289],
                    [60.08463455818353, 44.90380671613201],
                    [60.077403974563175, 45.046212550598135],
                    [59.87655568675822, 45.040891215906676],
                    ...
                  ]
                ],
                'type': 'Polygon'
              },
              'properties': {
                'cs_code': 'EPSG:32641',
                'key': '1024:16:15.0:41:-16:324',
                'outputBounds': [254000.0, 4976400.0, 269840.0, 4992240.0],
                'pad': 16,
                'resolution': 15.0,
                'ti': -16,
                'tilesize': 1024,
                'tj': 324,
                'zone': 41
              },
              'type': 'Feature'
            }
        """
        params = {
            'resolution': resolution,
            'tilesize': tilesize,
            'pad': pad,
        }

        r = self.session.get('/dlkeys/from_latlon/%f/%f' % (lat, lon), params=params)

        return DotDict(r.json())

    def dltile(self, key):
        """
        Given a DLTile key, return a DLTile GeoJSON Feature

        :param str key: A DLTile key that identifies a DLTile

        :return: A DLTile GeoJSON Feature

        Example::

            >>> from descarteslabs.client.services import Raster
            >>> Raster().dltile("1024:16:15.0:41:-16:324")
            {
              'geometry': {
                'coordinates': [
                  [
                    [59.88428127486419, 44.89851158847289],
                    [60.08463455818353, 44.90380671613201],
                    [60.077403974563175, 45.046212550598135],
                    [59.87655568675822, 45.040891215906676],
                    ...
                  ]
                ],
                'type': 'Polygon'
              },
              'properties': {
                'cs_code': 'EPSG:32641',
                'key': '1024:16:15.0:41:-16:324',
                'outputBounds': [254000.0, 4976400.0, 269840.0, 4992240.0],
                'pad': 16,
                'resolution': 15.0,
                'ti': -16,
                'tilesize': 1024,
                'tj': 324,
                'zone': 41
              },
              'type': 'Feature'
            }
        """

        r = self.session.get('/dlkeys/%s' % key)

        return DotDict(r.json())

    def raster(
            self,
            inputs,
            bands=None,
            scales=None,
            data_type=None,
            output_format='GTiff',
            srs=None,
            dimensions=None,
            resolution=None,
            bounds=None,
            bounds_srs=None,
            cutline=None,
            place=None,
            align_pixels=False,
            resampler=None,
            dltile=None,
            save=False,
            outfile_basename=None,
            **pass_through_params
    ):
        """Given a list of :class:`Metadata <descarteslabs.services.Metadata>` identifiers,
        retrieve a translated and warped mosaic as an image file.

        :param inputs: List of :class:`Metadata` identifiers.
        :param bands: List of requested bands. If the last item in the list is an alpha
            band (with data range `[0, 1]`) it affects rastering of all other bands:
            When rastering multiple images, they are combined image-by-image only where
            each respective image's alpha band is `1` (pixels where the alpha band is not
            `1` are "transparent" in the overlap between images). If a pixel is fully
            masked considering all combined alpha bands it will be `0` in all non-alpha
            bands.
        :param scales: List of tuples specifying the scaling to be applied to each band.
            A tuple has 4 elements in the order ``(src_min, src_max, out_min, out_max)``,
            meaning values in the source range ``src_min`` to ``src_max`` will be scaled
            to the output range ``out_min`` to ``out_max``. A tuple with 2 elements
            ``(src_min, src_max)`` is also allowed, in which case the output range
            defaults to ``(0, 255)`` (a useful default for the common output type
            ``Byte``).  If no scaling is desired for a band, use ``None``.  This tuple
            format and behaviour is identical to GDAL's scales during translation.
            Example argument: ``[(0, 10000, 0, 127), None, (0, 10000)]`` - the first
            band will have source values 0-10000 scaled to 0-127, the second band will
            not be scaled, the third band will have 0-10000 scaled to 0-255.
        :param str output_format: Output format (one of ``GTiff``, ``PNG``, ``JPEG``).
        :param str data_type: Output data type (one of ``Byte``, ``UInt16``, ``Int16``,
            ``UInt32``, ``Int32``, ``Float32``, ``Float64``).
        :param str srs: Output spatial reference system definition understood by GDAL.
        :param float resolution: Desired resolution in output SRS units. Incompatible with
            `dimensions`
        :param tuple dimensions: Desired output (width, height) in pixels. Incompatible with
            `resolution`
        :param str cutline: A GeoJSON feature or geometry to be used as a cutline.
        :param str place: A slug identifier to be used as a cutline.
        :param tuple bounds: ``(min_x, min_y, max_x, max_y)`` in target SRS.
        :param str bounds_srs:
            Override the coordinate system in which bounds are expressed.
            If not given, bounds are assumed to be expressed in the output SRS.
        :param bool align_pixels: Align pixels to the target coordinate system.
        :param str resampler: Resampling algorithm to be used during warping (``near``,
            ``bilinear``, ``cubic``, ``cubicsplice``, ``lanczos``, ``average``, ``mode``,
            ``max``, ``min``, ``med``, ``q1``, ``q3``).
        :param str dltile: a dltile key used to specify the resolution, bounds, and srs.
        :param bool save: Write resulting files to disk. Default: False
        :param str outfile_basename: If 'save' is True, override default filename using
            this string as a base.

        :return: A dictionary with two keys, ``files`` and ``metadata``. The value for
            ``files`` is a dictionary mapping file names to binary data for files (at the
            moment there will always be only a single file with the appropriate file
            extension based on the ``output_format`` requested). The value for ``metadata``
            is a dictionary containing details about the raster operation that happened.
            These details can be useful for debugging but shouldn't otherwise be relied on
            (there are no guarantees that certain keys will be present).
        """
        cutline = as_json_string(cutline)

        if place:
            places = Places()
            places.auth = self.auth
            shape = places.shape(place, geom='low')
            cutline = json.dumps(shape['geometry'])

        params = {
            'ids': inputs,
            'bands': bands,
            'scales': scales,
            'ot': data_type,
            'of': output_format,
            'srs': srs,
            'resolution': resolution,
            'shape': cutline,
            'outputBounds': bounds,
            'outputBoundsSRS': bounds_srs,
            'outsize': dimensions,
            'targetAlignedPixels': align_pixels,
            'resampleAlg': resampler,
        }
        params.update(pass_through_params)

        if dltile is not None:
            if isinstance(dltile, dict):
                params['dltile'] = dltile['properties']['key']
            else:
                params['dltile'] = dltile

        r = self.session.post('/raster', json=params)

        raw = BytesIO(r.content)

        json_resp = json.loads(raw.readline().decode('utf-8').strip())

        num_files = json_resp['files']
        json_resp['files'] = {}

        for _ in range(num_files):
            file_meta = json.loads(raw.readline().decode('utf-8').strip())

            fn = file_meta['name']
            data = raw.read(file_meta['length'])

            if outfile_basename:
                outfilename = "{}.{}".format(
                    outfile_basename,
                    ".".join(os.path.basename(fn).split(".")[1:])
                )
            else:
                outfilename = fn

            json_resp['files'][outfilename] = data

        if save:
            for filename, data in six.iteritems(json_resp['files']):
                with open(filename, "wb") as f:
                    f.write(data)

        return DotDict(json_resp)

    def ndarray(
            self,
            inputs,
            bands=None,
            scales=None,
            data_type=None,
            srs=None,
            resolution=None,
            dimensions=None,
            cutline=None,
            place=None,
            bounds=None,
            bounds_srs=None,
            align_pixels=False,
            resampler=None,
            order='image',
            dltile=None,
            **pass_through_params
    ):
        """Retrieve a raster as a NumPy array.

        :param inputs: List of :class:`Metadata` identifiers.
        :param bands: List of requested bands. If the last item in the list is an alpha
            band (with data range `[0, 1]`) it affects rastering of all other bands:
            When rastering multiple images, they are combined image-by-image only where
            each respective image's alpha band is `1` (pixels where the alpha band is not
            `1` are "transparent" in the overlap between images). If a pixel is fully
            masked considering all combined alpha bands it will be `0` in all non-alpha
            bands.
        :param scales: List of tuples specifying the scaling to be applied to each band.
            A tuple has 4 elements in the order ``(src_min, src_max, out_min, out_max)``,
            meaning values in the source range ``src_min`` to ``src_max`` will be scaled
            to the output range ``out_min`` to ``out_max``. A tuple with 2 elements
            ``(src_min, src_max)`` is also allowed, in which case the output range
            defaults to ``(0, 255)`` (a useful default for the common output type
            ``Byte``).  If no scaling is desired for a band, use ``None``. This tuple
            format and behaviour is identical to GDAL's scales during translation.
            Example argument: ``[(0, 10000, 0, 127), None, (0, 10000)]`` - the first
            band will have source values 0-10000 scaled to 0-127, the second band will
            not be scaled, the third band will have 0-10000 scaled to 0-255.
        :param str data_type: Output data type (one of ``Byte``, ``UInt16``, ``Int16``,
            ``UInt32``, ``Int32``, ``Float32``, ``Float64``).
        :param str srs: Output spatial reference system definition understood by GDAL.
        :param float resolution: Desired resolution in output SRS units. Incompatible with
            `dimensions`
        :param tuple dimensions: Desired output (width, height) in pixels. Incompatible with
            `resolution`
        :param str cutline: A GeoJSON feature or geometry to be used as a cutline.
        :param str place: A slug identifier to be used as a cutline.
        :param tuple bounds: ``(min_x, min_y, max_x, max_y)`` in target SRS.
        :param str bounds_srs:
            Override the coordinate system in which bounds are expressed.
            If not given, bounds are assumed to be expressed in the output SRS.
        :param bool align_pixels: Align pixels to the target coordinate system.
        :param str resampler: Resampling algorithm to be used during warping (``near``,
            ``bilinear``, ``cubic``, ``cubicsplice``, ``lanczos``, ``average``, ``mode``,
            ``max``, ``min``, ``med``, ``q1``, ``q3``).
        :param str order: Order of the returned array. `image` returns arrays as
            ``(row, column, band)`` while `gdal` returns arrays as ``(band, row, column)``.
        :param str dltile: a dltile key used to specify the resolution, bounds, and srs.

        :return: A tuple of ``(np_array, metadata)``. The first element (``np_array``) is
            the rastered scene as a NumPy array. The second element (``metadata``) is a
            dictionary containing details about the raster operation that happened. These
            details can be useful for debugging but shouldn't otherwise be relied on (there
            are no guarantees that certain keys will be present).
        """
        cutline = as_json_string(cutline)

        if place is not None:
            places = Places(auth=self.auth)
            shape = places.shape(place, geom='low')
            cutline = json.dumps(shape['geometry'])

        params = {
            'ids': inputs,
            'bands': bands,
            'scales': scales,
            'ot': data_type,
            'srs': srs,
            'resolution': resolution,
            'shape': cutline,
            'outputBounds': bounds,
            'outputBoundsSRS': bounds_srs,
            'outsize': dimensions,
            'targetAlignedPixels': align_pixels,
            'resampleAlg': resampler,
        }
        params.update(pass_through_params)

        if dltile is not None:
            if isinstance(dltile, dict):
                params['dltile'] = dltile['properties']['key']
            else:
                params['dltile'] = dltile

        can_blosc = not isinstance(blosc, ThirdParty)

        if can_blosc:
            params['of'] = 'blosc'
        else:
            params['of'] = 'npz'

        r = self.session.post('/npz', json=params, stream=True)

        if can_blosc:
            metadata = json.loads(r.raw.readline().decode('utf-8').strip())
            array_meta = json.loads(r.raw.readline().decode('utf-8').strip())
            array = read_blosc_array(array_meta, r.raw)
        else:
            npz = np.load(BytesIO(r.content))
            array = npz['data']
            metadata = json.loads(npz['metadata'].tostring().decode('utf-8'))

        if len(array.shape) > 2:
            if order == 'image':
                return array.transpose((1, 2, 0)), metadata
            elif order == 'gdal':
                return array, metadata
        else:
            return array, DotDict(metadata)

    def _serial_ndarray(self, id_groups, *args, **kwargs):
        for i, id_group in enumerate(id_groups):
            arr, meta = self.ndarray(id_group, *args, **kwargs)
            yield i, arr, meta

    def _threaded_ndarray(self, id_groups, *args, **kwargs):
        """
        Thread ndarray calls by id group, keeping the same `args` and
        `kwargs` for each raster.ndarray call.
        """
        max_workers = kwargs.pop(
            "max_workers",
            min(len(id_groups), DEFAULT_MAX_WORKERS)
        )
        try:
            futures = concurrent.futures
        except ImportError:
            logging.warning(
                "Failed to import concurrent.futures. ndarray calls will be serial"
            )
            # Fallback to serial execution
            for i, arr, meta in self._serial_ndarray(id_groups, *args, **kwargs):
                yield i, arr, meta
            return

        with futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
            future_ndarrays = {}
            for i, id_group in enumerate(id_groups):
                future_ndarrays[executor.submit(
                    self.ndarray, id_group, *args, **kwargs)
                ] = i

            for future in futures.as_completed(future_ndarrays):
                i = future_ndarrays[future]
                arr, meta = future.result()
                yield i, arr, meta

    def stack(
            self,
            inputs,
            bands=None,
            scales=None,
            data_type="UInt16",
            srs=None,
            resolution=None,
            dimensions=None,
            cutline=None,
            place=None,
            bounds=None,
            bounds_srs=None,
            align_pixels=False,
            resampler=None,
            order='image',
            dltile=None,
            max_workers=None,
            **pass_through_params
    ):
        """Retrieve a stack of rasters as a 4-D NumPy array.

        To ensure every raster in the stack has the same shape and covers the same
        spatial extent, you must either:
        * set ``dltile``, or
        * set [``resolution`` or ``dimensions``], ``srs``, and ``bounds``

        :param inputs: List, or list of lists, of :class:`Metadata` identifiers.
            The stack will follow the same order as this list.
            Each element in the list is treated as a separate input to ``raster.ndarray``,
            so if a list of lists is given, each sublist's identifiers will be mosaiced together
            to become a single level in the stack.
        :param bands: List of requested bands. If the last item in the list is an alpha
            band (with data range `[0, 1]`) it affects rastering of all other bands:
            When rastering multiple images, they are combined image-by-image only where
            each respective image's alpha band is `1` (pixels where the alpha band is not
            `1` are "transparent" in the overlap between images). If a pixel is fully
            masked considering all combined alpha bands it will be `0` in all non-alpha
            bands.
        :param scales: List of tuples specifying the scaling to be applied to each band.
            A tuple has 4 elements in the order ``(src_min, src_max, out_min, out_max)``,
            meaning values in the source range ``src_min`` to ``src_max`` will be scaled
            to the output range ``out_min`` to ``out_max``. A tuple with 2 elements
            ``(src_min, src_max)`` is also allowed, in which case the output range
            defaults to ``(0, 255)`` (a useful default for the common output type
            ``Byte``).  If no scaling is desired for a band, use ``None``. This tuple
            format and behaviour is identical to GDAL's scales during translation.
            Example argument: ``[(0, 10000, 0, 127), None, (0, 10000)]`` - the first
            band will have source values 0-10000 scaled to 0-127, the second band will
            not be scaled, the third band will have 0-10000 scaled to 0-255.
        :param str data_type: Output data type (one of ``Byte``, ``UInt16``, ``Int16``,
            ``UInt32``, ``Int32``, ``Float32``, ``Float64``).
        :param str srs: Output spatial reference system definition understood by GDAL.
        :param float resolution: Desired resolution in output SRS units. Incompatible with
            `dimensions`
        :param tuple dimensions: Desired output (width, height) in pixels. Incompatible with
            `resolution`
        :param str cutline: A GeoJSON feature or geometry to be used as a cutline.
        :param str place: A slug identifier to be used as a cutline.
        :param tuple bounds: ``(min_x, min_y, max_x, max_y)`` in target SRS.
        :param str bounds_srs:
            Override the coordinate system in which bounds are expressed.
            If not given, bounds are assumed to be expressed in the output SRS.
        :param bool align_pixels: Align pixels to the target coordinate system.
        :param str resampler: Resampling algorithm to be used during warping (``near``,
            ``bilinear``, ``cubic``, ``cubicsplice``, ``lanczos``, ``average``, ``mode``,
            ``max``, ``min``, ``med``, ``q1``, ``q3``).
        :param str order: Order of the returned array. `image` returns arrays as
            ``(scene, row, column, band)`` while `gdal` returns arrays as ``(scene, band, row, column)``.
        :param str dltile: a dltile key used to specify the resolution, bounds, and srs.
        :param int max_workers: Maximum number of threads over which to
            parallelize individual ndarray calls. If `None`, will be set to the minimum
            of the number of inputs and `DEFAULT_MAX_WORKERS`.

        :return: A tuple of ``(stack, metadata)``.
            ``stack``: 4D ndarray. The axes are ordered ``(scene, band, y, x)``
            (or ``(scene, y, x, band)`` if ``order="gdal"``). The scenes in the outermost
            axis are in the same order as the list of identifiers given as ``inputs``.
            ``metadata``: List[dict] of the rasterization metadata for each element in ``inputs``.
            As with the metadata returned by :meth:`ndarray` and :meth:`raster`, these dictionaries
            contain useful information about the raster, such as its geotransform matrix and WKT
            of its coordinate system, but there are no guarantees that certain keys will be present.
        """
        if not isinstance(inputs, (list, tuple)):
            raise TypeError("Inputs must be a list or tuple, instead got '{}'".format(type(inputs)))

        params = dict(
            bands=bands,
            scales=scales,
            data_type=data_type,
            srs=srs,
            resolution=resolution,
            dimensions=dimensions,
            cutline=cutline,
            place=place,
            bounds=bounds,
            bounds_srs=bounds_srs,
            align_pixels=align_pixels,
            resampler=resampler,
            order=order,
            dltile=dltile,
            max_workers=max_workers,
            **pass_through_params
        )

        if dltile is None:
            if resolution is None and dimensions is None:
                raise ValueError("Must set `resolution` or `dimensions`")
            if srs is None:
                raise ValueError("Must set `srs`")
            if bounds is None:
                raise ValueError("Must set `bounds`")

        full_stack = None
        metadata = [None] * len(inputs)
        for i, arr, meta in self._threaded_ndarray(inputs, **params):
            if len(arr.shape) == 2:
                if order == "image":
                    arr = np.expand_dims(arr, -1)
                elif order == "gdal":
                    arr = np.expand_dims(arr, 0)
                else:
                    raise ValueError("Unknown order '{}'; should be one of 'image' or 'gdal'".format(order))
            if full_stack is None:
                stack_shape = (len(inputs),) + arr.shape
                full_stack = np.empty(stack_shape, dtype=arr.dtype)
            full_stack[i] = arr
            metadata[i] = meta

        return full_stack, metadata



# Copyright 2018 Descartes Labs.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from descarteslabs.client.services.places import Places


class TestPlaces(unittest.TestCase):
    instance = None

    @classmethod
    def setUpClass(cls):
        cls.instance = Places()

    def test_placetypes(self):
        data = self.instance.placetypes()
        self.assertGreaterEqual(10, len(data))

    def test_find(self):
        r = self.instance.find('united-states_iowa')
        self.assertEqual(1, len(r))

    def test_search(self):
        r = self.instance.search('texas')
        self.assertEqual(8, len(r))

        r = self.instance.search('texas', country='united-states')
        self.assertEqual(7, len(r))

        r = self.instance.search('texas', country='united-states', placetype='county')
        self.assertEqual(2, len(r))

        r = self.instance.search('texas', country='united-states', region='oklahoma', placetype='county')
        self.assertEqual(1, len(r))

    def test_shape(self):
        r = self.instance.shape('north-america_united-states_iowa')
        self.assertEqual(85688713, r['id'])

    def test_prefix(self):
        # counties by default
        r = self.instance.prefix('north-america_united-states_iowa')
        self.assertEqual(99, len(r['features']))

        r = self.instance.prefix('north-america_united-states_iowa', placetype='district')
        self.assertEqual(9, len(r['features']))

    def test_sources(self):
        r = self.instance.sources()
        self.assertIn({'name': 'nass'}, r)

    def test_categories(self):
        r = self.instance.categories()
        self.assertIn({'name': 'corn'}, r)

    def test_metrics(self):
        r = self.instance.metrics()
        self.assertIn({'name': 'yield', 'units': 'bu/ac'}, r)

    def test_value(self):
        r = self.instance.value('north-america_united-states', source='nass', category='corn', metric='yield')
        self.assertEqual(1, len(r))

    def test_statistics(self):
        r = self.instance.statistics('north-america_united-states', source='nass', category='corn', metric='yield',
                                     units='bu/ac')
        self.assertEqual(36, len(r))

    def test_data(self):
        r = self.instance.data('north-america_united-states', source='nass', category='corn', metric='yield',
                               date='2015-01-01', units='bu/ac')
        self.assertEqual(1439, len(r))


if __name__ == '__main__':
    unittest.main()



# Copyright 2018 Descartes Labs.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import unittest
import sys
from mock import patch
from tempfile import NamedTemporaryFile
import responses

from descarteslabs.client.auth import Auth
from descarteslabs.client.services.catalog import Catalog


@patch.object(Auth, 'token', 'token')
@patch.object(Auth, 'namespace', 'foo')
class TestCatalog(unittest.TestCase):
    instance = None

    def setUp(self):
        self.instance = Catalog()

    @staticmethod
    def validate_ndarray_callback(request):
        np.load(request.body)
        return (200, {}, '')

    @patch('descarteslabs.client.services.catalog.Catalog._do_upload', return_value=(False,))
    def test_upload_ndarray_dtype(self, _do_upload):
        unsupported_dtypes = ['uint64']

        for dtype in unsupported_dtypes:
            self.assertRaises(TypeError,
                              self.instance.upload_ndarray,
                              np.zeros((10, 10), dtype=dtype), 'product', 'key')

        for dtype in Catalog.UPLOAD_NDARRAY_SUPPORTED_DTYPES:
            self.instance.upload_ndarray(np.zeros((10, 10), dtype=dtype), 'product', 'key')

    @unittest.skipIf(sys.version_info.major == 3, "Test only makes sense in py2")
    def test_upload_image_deprecated_file_type(self):
        # in py2 NamedTemporaryFile produces a file object not a IOBase object
        with NamedTemporaryFile() as tmp:
            self.assertRaises(
                Exception,
                self.instance.upload_image, tmp, 'product'
            )

    def test_upload_image_bad_path(self):
        name = None
        with NamedTemporaryFile() as tmp:
            name = tmp.file
        self.assertRaises(
            Exception,
            self.instance.upload_image, name, 'product'
        )

    def test_upload_image_multi_file_no_list(self):
        with NamedTemporaryFile() as tmp:
            self.assertRaises(
                ValueError,
                self.instance.upload_image, tmp.name, 'product', multi=True
            )

    def test_upload_image_multi_file_no_image_id(self):
        with NamedTemporaryFile() as tmp:
            self.assertRaises(
                ValueError,
                self.instance.upload_image, [tmp.name, tmp.name], 'product', multi=True
            )

    @responses.activate
    def test_upload_image(self):
        product = 'foo:product_id'
        gcs_upload_url = 'https://gcs_upload_url.com'
        upload_url = 'https://platform.descarteslabs.com/metadata/v1/catalog/products/{}/images/upload/{}'
        with NamedTemporaryFile() as tmp:
            tmp.write(b'foo')
            responses.add(
                responses.POST,
                upload_url.format(product, tmp.name.split('/')[-1]),
                body=gcs_upload_url
            )
            responses.add(responses.PUT, gcs_upload_url)
            self.instance.upload_image(tmp.name, product)

    @responses.activate
    def test_upload_ndarray(self):
        product = 'foo:product_id'
        gcs_upload_url = 'https://gcs_upload_url.com'
        upload_url = 'https://platform.descarteslabs.com/metadata/v1/catalog/products/{}/images/upload/key'
        responses.add(
            responses.POST,
            upload_url.format(product),
            body=gcs_upload_url,
        )
        responses.add_callback(responses.PUT, gcs_upload_url, callback=self.validate_ndarray_callback)
        self.instance.upload_ndarray(np.zeros((10, 10)), product, 'key')


if __name__ == '__main__':
    unittest.main()



import six
from six.moves import reprlib
from itertools import islice


class DotDict(dict):
    """
    Subclass of dict, with "dot" (attribute) access to keys,
    a pretty-printed repr, which indents and truncates large containers,
    and a JSON repr for Jupyter Lab.

    Any dicts stored in DotDict are returned as DotDicts, to allow chained attribute access.
    Any lists stored in DotDict are returned as DotLists, which return any contained dict items as DotDicts,
    allowing chained attribute access past list indexing.

    The repr() of a DotDict is truncated for readability, but str() is not.

    Example
    -------

    >>> d = DotDict(a=1, b=[{"foo": "bar"}])
    >>> d.a
        1
    >>> d["a"]
        1
    >>> d.b
        [
          {
            'foo': 'bar'
          }
        ]
    >>> d.b[0].foo
        'bar'
    """
    __slots__ = ()  # no need for a namespace __dict__ when DotDict is a dict already

    def _repr_json_(self):
        return self, {'expanded': False}

    @classmethod
    def _box(cls, value):
        "If value is a dict or list, return it as a DotDict or DotList, otherwise return unmodified value."
        if type(value) is dict:
            return cls(value)
        elif type(value) is list:
            return DotList(value)
        else:
            return value

    def __getitem__(self, key):
        """
        x.__getitem__(y) <==> x[y]
        If x[y] is a dict or list, it is returned as a DotDict or DotList.
        """
        try:
            v = dict.__getitem__(self, key)
        except KeyError:
            six.raise_from(KeyError(key), None)

        v = self._box(v)
        self[key] = v
        return v

    def __getattr__(self, attr):
        """
        self.attr <==> self[attr]
        If x[y] is a dict or list, it is returned as a DotDict or DotList.
        """
        try:
            return self[attr]
        except KeyError:
            try:
                return object.__getattribute__(self, attr)
            except AttributeError:
                six.raise_from(AttributeError(attr), None)

    def __setattr__(self, attr, val):
        "self.attr = x <==> self[attr] = x"
        self[attr] = val

    def __delattr__(self, attr):
        "del self.attr <==> del self[attr]"
        try:
            del self[attr]
        except KeyError:
            six.raise_from(AttributeError(attr), None)

    def __dir__(self):
        return list(self.keys()) + dir(dict)

    def __repr__(self):
        return idr.repr(self)

    def __str__(self):
        return untruncated_idr.repr(self)

    def items(self):
        """
        Equivalent to dict.items in this Python version.
        Values that are plain dicts or lists are returned as DotDicts or DotLists.
        """
        if six.PY2:
            return list(self.iteritems())
        else:
            return DotDict_items(self)

    def iteritems(self):
        """
        D.iteritems() -> an iterator over the (key, value) items of D
        Values that are dicts or lists are cast to DotDicts and DotLists.
        Only implemented in Python 2.
        """
        if six.PY2:
            for k, v in dict.iteritems(self):
                boxed = self._box(v)
                if boxed is not v:
                    self[k] = boxed
                yield (k, boxed)
        else:
            raise AttributeError("'iteritems' is only implemented in Python 2; use .items()")

    def values(self):
        """
        Equivalent to dict.values in this Python version.
        Values that are plain dicts or lists are returned as DotDicts or DotLists.
        """
        if six.PY2:
            return list(self.itervalues())
        else:
            return DotDict_values(self)

    def itervalues(self):
        """
        D.itervalues() -> an iterator over the values of D
        Values that are dicts or lists are cast to DotDicts and DotLists.
        Only implemented in Python 2.
        """
        if six.PY2:
            for k, v in self.iteritems():
                yield v
        else:
            raise AttributeError("'itervalues' is only implemented in Python 2; use .values()")

    def get(self, key, default=None):
        """
        D.get(k[,d]) -> D[k] if k in D, else d.  d defaults to None.
        Values that are dicts or lists are cast to DotDicts and DotLists.
        """
        try:
            return self[key]
        except KeyError:
            return self._box(default)

    def pop(self, key, default=None):
        """
        D.pop(k[,d]) -> v, remove specified key and return the corresponding value.
        If key is not found, d is returned if given, otherwise KeyError is raised.
        If v is a dict or list, it is returned as a DotDict or DotList.
        """
        result = dict.pop(self, key, default)
        return self._box(result)

    def popitem(self):
        """
        D.popitem() -> (k, v), remove and return some (key, value) pair as a
        2-tuple; but raise KeyError if D is empty.
        If v is a dict or list, it is returned as a DotDict or DotList.
        """
        k, v = dict.popitem(self)
        return k, self._box(v)

    def setdefault(self, key, default=None):
        """
        D.setdefault(k[,d]) -> D.get(k,d), also set D[k]=d if k not in D
        If d is a dict or list, it is returned as a DotDict or DotList.
        """
        try:
            return self[key]
        except KeyError:
            default = self._box(default)
            self[key] = default
            return default

    def asdict(self):
        """
        D.asdict() -> a deep copy of D, where any DotDicts or DotLists contained are converted to plain types.
        Raises RuntimeError if the container is recursive (contains itself as a value).
        """
        # TODO: does not handle recursive structures

        # note: we're assuming here that any plain dict/list doesn't contain Dot-types within it.
        # this is safe for normal usage of DotDict: any assignment to a DotDict will cause all the levels
        # in the hierarchy to be converted to Dot-types
        # However, if someone creates a plain dict, assigns DotDicts as its values, then assigns
        # *that* plain dict to a value in a DotDict, the asdict of the containing DotDict will stop when
        # it hits the plain dict.
        # The most probable case here is assigning Dot-types as values in a dictionary or list comprehension.

        unboxed = {}
        iterator = dict.iteritems if six.PY2 else dict.items
        for k, v in iterator(self):
            if isinstance(v, DotDict):
                v = v.asdict()
            if isinstance(v, DotList):
                v = v.aslist()
            unboxed[k] = v
        return unboxed


class DotDict_view(object):
    """Wrapper around a dictionary view object that yields dicts and lists as DotDicts and DotLists when iterated."""
    __slots__ = ("_view",)

    def __init__(self, dotdict):
        self._view = dict.items(dotdict)
        self._dict = dotdict

    def __iter__(self):
        """Implement iter(self)."""
        for k, v in self._view:
            boxed = DotDict._box(v)
            if boxed is not v:
                self._dict[k] = boxed
            yield k, boxed

    def __len__(self):
        """Return len(self)."""
        return self._view.__len__()

    def __repr__(self):
        """Return repr(self)."""
        return "{}({})".format(self.__class__.__name__, list(self))


class DotDict_values(DotDict_view):
    """Wrapper around a dict_values object that yields dicts and lists as DotDicts and DotLists when iterated."""

    def __iter__(self):
        for k, v in super(DotDict_values, self).__iter__():
            yield v


class DotDict_items(DotDict_view):
    """Wrapper around a dict_values object that yields dicts and lists as DotDicts and DotLists when iterated."""

    def __and__(self, value):
        """Return self&value."""
        return self._view.__and__(value)

    def __contains__(self, key):
        """Return key in self."""
        return self._view.__contains__(key)

    def __eq__(self, value):
        """Return self==value."""
        return self._view.__eq__(value)

    def __ge__(self, value):
        """Return self>=value."""
        return self._view.__ge__(value)

    def __gt__(self, value):
        """Return self>value."""
        return self._view.__gt__(value)

    def __le__(self, value):
        """Return self<=value."""
        return self._view.__le__(value)

    def __lt__(self, value):
        """Return self<value."""
        return self._view.__lt__(value)

    def __ne__(self, value):
        """Return self!=value."""
        return self._view.__ne__(value)

    def __or__(self, value):
        """Return self|value."""
        return self._view.__or__(value)

    def __rand__(self, value):
        """Return value&self."""
        return self._view.__rand__(value)

    def __ror__(self, value):
        """Return value|self."""
        return self._view.__ror__(value)

    def __rsub__(self, value):
        """Return value-self."""
        return self._view.__rsub__(value)

    def __rxor__(self, value):
        """Return value^self."""
        return self._view.__rxor__(value)

    def __sub__(self, value):
        """Return self-value."""
        return self._view.__sub__(value)

    def __xor__(self, value):
        """Return self^value."""
        return self._view.__xor__(value)

    def isdisjoint(self, iterable):
        """Return True if the view and the given iterable have a null intersection."""
        return self._view.isdisjoint(iterable)


class DotList(list):
    """
    Returns contained dicts as DotDicts (and contained lists as DotLists),
    soley to allow attribute access past list indexing
    """
    __slots__ = ()

    def __getitem__(self, i):
        """
        x.__getitem__(y) <==> x[y]
        If x[y] is a dict or list, it is returned as a DotDict or DotList.
        """
        try:
            item = list.__getitem__(self, i)
        except IndexError:
            six.raise_from(IndexError("list index out of range"), None)

        item = DotDict._box(item)
        self[i] = item
        return item

    def __getslice__(self, i, j):
        return self.__getitem__(slice(i, j))

    def __iter__(self):
        for i in range(0, len(self)):
            yield self[i]

    def __repr__(self):
        return idr.repr(self)

    def __str__(self):
        return untruncated_idr.repr(self)

    def pop(self, i=-1):
        """
        L.pop([index]) -> item -- remove and return item at index (default last).
        If item is a dict or list, it is returned as a DotDict or DotList.
        Raises IndexError if list is empty or index is out of range.
        """
        result = list.pop(self, i)
        return DotDict._box(result)

    def aslist(self):
        """
        L.aslist() -> a deep copy of L, where any DotDicts or DotLists contained are converted to plain types.
        Raises RuntimeError if the container is recursive (contains itself as a value).
        """
        unboxed = list(self)
        for i, obj in enumerate(unboxed):
            if isinstance(obj, DotList):
                unboxed[i] = obj.aslist()
            elif isinstance(obj, DotDict):
                unboxed[i] = obj.asdict()
        return unboxed


def _possibly_sorted(x):
    # Since not all sequences of items can be sorted and comparison
    # functions may raise arbitrary exceptions, return an unsorted
    # sequence in that case.
    try:
        return sorted(x)
    except Exception:
        return list(x)


class IndentedRepr(reprlib.Repr, object):

    def __init__(self):
        super(IndentedRepr, self).__init__()

        self.maxstring = 90  # about the maximum width of a Jupyter Notebook
        self.maxlevel = 6
        self.maxlist = 4
        self.maxdict = None

        self.indent = 2

    def repr_DotDict(self, x, level):
        return self.repr_dict(x, level)

    def repr_DotList(self, x, level):
        return self.repr_list(x, level)

    def repr_unicode(self, x, level):
        return self.repr_str(x, level)

    def repr1(self, x, level):
        # repr1 is explicity defined rather than inherited,
        # because py2 and py3 have different implementations---py2 inlines repr_instance, basically
        typename = type(x).__name__
        if ' ' in typename:
            parts = typename.split()
            typename = '_'.join(parts)
        if hasattr(self, 'repr_' + typename):
            return getattr(self, 'repr_' + typename)(x, level)
        else:
            return self.repr_instance(x, level)

    def repr_dict(self, x, level):
        n = len(x)
        if n == 0:
            return '{}'
        if self.maxlevel is not None:
            depth = self.maxlevel - level
            if level <= 0:
                return '{...}'
            newlevel = level - 1
        else:
            if level is None:
                level = 0
            depth = level
            newlevel = level + 1
        repr1 = self.repr1
        pieces = []
        for key in islice(_possibly_sorted(x), self.maxdict if self.maxdict is not None and depth > 0 else None):
            keyrepr = repr1(key, newlevel)
            valrepr = repr1(x[key], newlevel)
            pieces.append('%s: %s' % (keyrepr, valrepr))
        if self.maxdict is not None and n > self.maxdict and depth > 0:
            pieces.append('...')

        outer_indent = ' ' * (self.indent * depth)
        inner_indent = outer_indent + ' ' * self.indent
        s = (',\n%s' % inner_indent).join(pieces)
        return '{\n%s%s\n%s}' % (inner_indent, s, outer_indent)

    def _repr_iterable(self, x, level, left, right, maxiter, trail=''):
        n = len(x)
        if self.maxlevel is not None:
            depth = self.maxlevel - level
            newlevel = level - 1
        else:
            if level is None:
                level = 0
            depth = level
            newlevel = level + 1

        outer_indent = ' ' * (self.indent * depth)
        inner_indent = outer_indent + ' ' * self.indent

        if self.maxlevel is not None and level <= 0 and n:
            s = '...'
        else:
            repr1 = self.repr1
            pieces = [repr1(elem, newlevel)
                      for elem in islice(x, maxiter if maxiter is not None and depth > 0 else None)]

            has_multiline_pieces = any('\n' in piece for piece in pieces)

            if maxiter is not None and n > maxiter and depth > 0:
                pieces.append('...')

            if has_multiline_pieces or maxiter is not None and n > maxiter:
                # multiline if long list, or components have line breaks (prevents weird closing bracket indentation)
                s = (',\n%s' % inner_indent).join(pieces)
                s = '\n%s%s\n%s' % (inner_indent, s, outer_indent)
            else:
                # single line if short
                s = ', '.join(pieces)

            if n == 1 and trail:
                right = trail + right

        return '%s%s%s' % (left, s, right)

    def repr_str(self, x, level):
        s = repr(x[:self.maxstring])
        if self.maxstring is not None:
            if len(s) > self.maxstring:
                i = max(0, (self.maxstring - 3) // 2)
                j = max(0, self.maxstring - 3 - i)
                s = repr(x[:i] + x[len(x) - j:])
                s = s[:i] + '...' + s[len(s) - j:]
        return s

    def repr_int(self, x, level):
        return self.repr_long(x, level)

    def repr_long(self, x, level):
        s = repr(x)  # XXX Hope this isn't too slow...
        if self.maxlong is not None and len(s) > self.maxlong:
            i = max(0, (self.maxlong - 3) // 2)
            j = max(0, self.maxlong - 3 - i)
            s = s[:i] + '...' + s[len(s) - j:]
        return s

    def repr_instance(self, x, level):
        try:
            s = repr(x)
            # Bugs in x.__repr__() can cause arbitrary
            # exceptions -- then make up something
        except Exception:
            return '<%s instance at %#x>' % (x.__class__.__name__, id(x))
        if self.maxother is not None and len(s) > self.maxother:
            i = max(0, (self.maxother - 3) // 2)
            j = max(0, self.maxother - 3 - i)
            s = s[:i] + '...' + s[len(s) - j:]
        return s


idr = IndentedRepr()
untruncated_idr = IndentedRepr()
untruncated_idr.maxlevel = None
untruncated_idr.maxdict = None
untruncated_idr.maxlist = None
untruncated_idr.maxtuple = None
untruncated_idr.maxset = None
untruncated_idr.maxfrozenset = None
untruncated_idr.maxdeque = None
untruncated_idr.maxarray = None
untruncated_idr.maxlong = None
untruncated_idr.maxstring = None
untruncated_idr.maxother = None



# Copyright 2018 Descartes Labs.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class ClientError(Exception):
    """ Base class """
    pass


class AuthError(ClientError):
    pass


class OauthError(AuthError):
    pass


class ServerError(Exception):
    status = 500


class BadRequestError(ServerError):
    status = 400


class NotFoundError(ServerError):
    status = 404


class ConflictError(ServerError):
    status = 409


class RateLimitError(ServerError):
    status = 429


class GatewayTimeoutError(ServerError):
    status = 504



# Copyright 2018 Descartes Labs.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random

import requests
from requests.adapters import HTTPAdapter
from warnings import warn
from urllib3.util.retry import Retry
from descarteslabs.client.auth import Auth
from descarteslabs.client.version import __version__
from descarteslabs.client.exceptions import ServerError, BadRequestError, NotFoundError, RateLimitError, \
    GatewayTimeoutError, ConflictError


class WrappedSession(requests.Session):
    def __init__(self, base_url, timeout=None):
        self.base_url = base_url
        self.timeout = timeout
        super(WrappedSession, self).__init__()

    def request(self, method, url, **kwargs):
        if self.timeout and 'timeout' not in kwargs:
            kwargs['timeout'] = self.timeout

        resp = super(WrappedSession, self).request(method, self.base_url + url, **kwargs)

        if resp.status_code >= 200 and resp.status_code < 400:
            return resp
        elif resp.status_code == 400:
            raise BadRequestError(resp.text)
        elif resp.status_code == 404:
            raise NotFoundError("404 %s %s" % (method, url))
        elif resp.status_code == 409:
            raise ConflictError(resp.text)
        elif resp.status_code == 429:
            raise RateLimitError(resp.text)
        elif resp.status_code == 504:
            raise GatewayTimeoutError(
                "Your request timed out on the server. "
                "Consider reducing the complexity of your request.")
        else:
            raise ServerError(resp.text)


class Service(object):
    TIMEOUT = (9.5, 30)

    RETRY_CONFIG = Retry(total=5,
                         read=2,
                         backoff_factor=random.uniform(1, 3),
                         method_whitelist=frozenset([
                             'HEAD', 'TRACE', 'GET', 'POST',
                             'PUT', 'OPTIONS', 'DELETE'
                         ]),
                         status_forcelist=[500, 502, 503, 504])

    ADAPTER = HTTPAdapter(max_retries=RETRY_CONFIG)

    def __init__(self, url, token=None, auth=None):
        if auth is None:
            auth = Auth()

        if token is not None:
            warn("setting token at service level will be removed in future", DeprecationWarning)
            auth._token = token

        self.auth = auth

        self.base_url = url

        self._session = self.build_session()

    @property
    def token(self):
        return self.auth.token

    @token.setter
    def token(self, token):
        self.auth._token = token

    @property
    def session(self):
        if self._session.headers.get('Authorization') != self.token:
            self._session.headers['Authorization'] = self.token

        return self._session

    def build_session(self):
        s = WrappedSession(self.base_url, timeout=self.TIMEOUT)
        s.mount('https://', self.ADAPTER)

        s.headers.update({
            "Content-Type": "application/json",
            "User-Agent": "dl-python/{}".format(__version__)
        })

        return s


class ThirdPartyService(object):
    TIMEOUT = (9.5, 30)

    RETRY_CONFIG = Retry(total=10,
                         read=2,
                         backoff_factor=random.uniform(1, 3),
                         method_whitelist=frozenset([
                             'HEAD', 'TRACE', 'GET', 'POST',
                             'PUT', 'OPTIONS', 'DELETE'
                         ]),
                         status_forcelist=[429, 500, 502, 503, 504])

    ADAPTER = HTTPAdapter(max_retries=RETRY_CONFIG)

    def __init__(self, url=''):
        self.base_url = url

        self._session = self.build_session()

    @property
    def session(self):
        return self._session

    def build_session(self):
        s = WrappedSession(self.base_url, timeout=self.TIMEOUT)
        s.mount('https://', self.ADAPTER)

        s.headers.update({
            "Content-Type": "application/octet-stream",
            "User-Agent": "dl-python/{}".format(__version__)
        })

        return s
